from typing import Optional, Dict, Any

import boto3
from aws_requests_auth.aws_auth import AWSRequestsAuth
from botocore.session import Session
from aws_requests_auth.boto_utils import BotoAWSRequestsAuth

from ai21.constants import BEDROCK_HOST_FORMAT, BEDROCK_URL_FORMAT
from ai21.errors import WrongInputTypeException
from ai21.http_client import HttpClient
from ai21.utils import get_global_configs, convert_to_ai21_object, get_value


class BedrockClient:
    def __init__(self, **params):
        self._http_client = self._build_http_client(params)

    def _build_http_client(self, params):
        global_configs = get_global_configs()
        headers = self._build_default_headers()
        passed_headers = params.get('headers', None)

        if passed_headers is not None:
            if not isinstance(passed_headers, Dict):
                raise WrongInputTypeException('headers', Dict, type(passed_headers))
            headers.update(passed_headers)

        timeout_sec = get_value('timeout_sec', params, global_configs, int)
        num_retries = get_value('num_retries', params, global_configs, int)

        return HttpClient(timeout_sec=timeout_sec, num_retries=num_retries, headers=headers)

    def _build_default_headers(self):
        headers = {'Content-Type': 'application/json'}

        return headers

    def execute_http_request(self, params: Optional[Dict[str, Any]], model_id: str,
                             boto_session: Optional[Session] = None):
        aws_region = self._get_aws_region(boto_session)

        auth_token = self._get_auth_token(boto_session)

        bedrock_url = f'{BEDROCK_URL_FORMAT.format(region_name=aws_region)}/model/{model_id}/invoke'

        response = self._http_client.execute_http_request(
            method='POST',
            url=bedrock_url,
            params=params,
            auth=auth_token,
        )

        return convert_to_ai21_object(response)

    def _get_aws_region(self, boto_session: Optional[boto3.Session]) -> str:
        if boto_session:
            return boto_session.region_name

        global_configs = get_global_configs()
        return global_configs.get('aws_region') or boto3.Session().region_name

    def _get_auth_token(self, boto_session: Optional[boto3.Session]) -> AWSRequestsAuth:
        aws_region = self._get_aws_region(boto_session)
        aws_host = BEDROCK_HOST_FORMAT.format(region_name=aws_region)
        aws_service = 'bedrock'

        if boto_session:
            credentials_obj = boto_session.get_credentials().get_frozen_credentials()

            return AWSRequestsAuth(
                aws_access_key=credentials_obj.access_key,
                aws_secret_access_key=credentials_obj.secret_key,
                aws_token=credentials_obj.token,
                aws_host=aws_host,
                aws_region=aws_region,
                aws_service=aws_service,
            )

        return BotoAWSRequestsAuth(
            aws_host=BEDROCK_HOST_FORMAT.format(region_name=aws_region),
            aws_region=aws_region,
            aws_service=aws_service,
        )



from ai21.ai21_object import AI21Object

from ai21.utils import get_global_configs

from ai21.ai21_studio_client import get_value


class AI21Module(AI21Object):

    @classmethod
    def get_base_url(cls, **params):
        global_configs = get_global_configs()
        api_host = get_value('api_host', params, global_configs, str)
        api_version = get_value('api_version', params, global_configs, str)
        return f'{api_host}/studio/{api_version}'

    @classmethod
    def get_module_url(cls, **params):
        return f'{cls.get_base_url(**params)}/{cls.MODULE_NAME}'







from dataclasses import dataclass
from typing import Optional


class Destination:
    pass


class AI21Destination(Destination):
    pass


@dataclass
class AWSDestination(Destination):
    pass


@dataclass
class BedrockDestination(AWSDestination):
    model_id: str
    boto_session: Optional["boto3.Session"] = None


@dataclass
class SageMakerDestination(AWSDestination):
    endpoint_name: str
    sm_session: Optional["sagemaker.session.Session"] = None



import json
from typing import Optional, List

from ai21.modules.destination import SageMakerDestination
from ai21.ai21_studio_client import AI21StudioClient
from ai21.utils import validate_mandatory_field, validate_unsupported_field


def execute_studio_request(task_url: str, params, method: str = 'POST'):
    client = AI21StudioClient(**params)
    return client.execute_http_request(method=method, url=task_url, params=params)


def execute_sm_request(destination: SageMakerDestination, params, mandatory_fields=None, unsupported_fields=None):
    call_name = f'Sagemaker {destination.endpoint_name}'

    if mandatory_fields is not None:
        for mandatory_field in mandatory_fields:
            validate_mandatory_field(mandatory_field, call_name, params)

    if unsupported_fields is not None:
        for unsupported_field in unsupported_fields:
            validate_unsupported_field(key=unsupported_field, call_name=call_name, params=params)

    input_json = json.dumps(params)

    from ai21.AWS_utils import invoke_sm_endpoint  # import here because boto3 exists only in AWS mode

    return invoke_sm_endpoint(
        endpoint_name=destination.endpoint_name,
        input_json=input_json,
        sm_session=destination.sm_session,
    )


def execute_bedrock_request(model_id: str, params, boto_session: Optional["boto3.Session"] = None,
                            unsupported_fields: Optional[List[str]] = None):
    call_name = 'Bedrock'

    if unsupported_fields is not None:
        for unsupported_field in unsupported_fields:
            validate_unsupported_field(key=unsupported_field, call_name=call_name, params=params)

    from ai21.bedrock_client import BedrockClient  # import here because boto3 exists only in AWS mode
    client = BedrockClient(**params)

    return client.execute_http_request(params=params, model_id=model_id, boto_session=boto_session)



from ai21.modules.resources.listable_resource import ListableResource
from ai21.modules.resources.retrievable_resource import RetrievableResource
from ai21.modules.resources.uploadable_resource import UploadableResource
from ai21.utils import validate_mandatory_field


class Dataset(UploadableResource, RetrievableResource, ListableResource):

    MODULE_NAME = 'dataset'

    @classmethod
    def upload(cls, file_path: str, **params):
        validate_mandatory_field(key='dataset_name', call_name=cls.MODULE_NAME, params=params, validate_type=True, expected_type=str)
        super().upload(file_path=file_path, file_param_name='dataset_file', **params)



from typing import Dict

from ai21.api_resources import supported_resources


def construct_ai21_object(obj):
    if not isinstance(obj, dict):
        return obj
    class_type = AI21Object
    type_key = obj.get("type", None)
    if type_key is not None:
        class_type = supported_resources().get(type_key, AI21Object)
    instance = class_type(obj)
    return instance


class AI21Object:

    def __init__(self, prop):
        self.values = prop

        if isinstance(prop, dict):
            for key, value in prop.items():
                if isinstance(value, dict):
                    setattr(self, key, construct_ai21_object(value))
                elif isinstance(value, list):
                    setattr(self, key, [construct_ai21_object(i) for i in value])
                else:
                    setattr(self, key, value)

    def __repr__(self):
        return str(self.values)

    def __getitem__(self, item):
        return getattr(self, item)




from ai21.constants import MODEL_ID_REMAPPING
from ai21.modules.destination import BedrockDestination, SageMakerDestination
from ai21.modules.resources.execution_utils import execute_sm_request, execute_bedrock_request, execute_studio_request
from ai21.modules.resources.nlp_task import NLPTask
from ai21.utils import validate_mandatory_field, log_warning


class Completion(NLPTask):
    MODULE_NAME = 'complete'

    @classmethod
    def _get_call_name(cls) -> str:
        return cls.MODULE_NAME

    @classmethod
    def _validate_params(cls, params):
        validate_mandatory_field(key='prompt', call_name=cls.MODULE_NAME, params=params, validate_type=True,
                                 expected_type=str)

        if params.get('stopSequences', None) is None:
            params['stopSequences'] = []

    @classmethod
    def _execute_bedrock_api(cls, destination: BedrockDestination, params):
        model_id = destination.model_id
        remap_model_id = MODEL_ID_REMAPPING.get(model_id)

        if remap_model_id:
            log_warning(f"The model ID: {model_id} is deprecated and will be soon removed. "
                      f"Please use {remap_model_id} instead")
            model_id = remap_model_id

        return execute_bedrock_request(
            model_id=model_id,
            boto_session=destination.boto_session,
            unsupported_fields=['model', 'custom_model'],
            params=params,
        )

    @classmethod
    def _execute_sm(cls, destination: SageMakerDestination, params):
        return execute_sm_request(
            destination=destination,
            unsupported_fields=['model', 'custom_model'],
            params=params,
        )

    @classmethod
    def _execute_studio_api(cls, params):
        validate_mandatory_field(
            key='model', call_name=cls._get_call_name(), params=params, validate_type=True,
            expected_type=str
        )

        model = params.get('model', None)
        custom_model = params.get('custom_model', None)
        experimental_mode = params.get('experimental_mode', False)

        if experimental_mode:
            model = f'experimental/{model}'

        url = f'{cls.get_base_url(**params)}/{model}'

        if custom_model is not None:
            url = f'{url}/{custom_model}'

        url = f'{url}/{cls.MODULE_NAME}'

        return execute_studio_request(task_url=url, params=params)



from ai21.constants import STUDIO_HOST, DEFAULT_API_VERSION, BedrockModelID
from ai21.modules.completion import Completion
from ai21.modules.dataset import Dataset
from ai21.modules.tokenization import Tokenization
from ai21.modules.custom_model import CustomModel
from ai21.modules.experimental import Experimental
from ai21.modules.paraphrase import Paraphrase
from ai21.modules.summarize import Summarize
from ai21.modules.summarize_by_segment import SummarizeBySegment
from ai21.modules.segmentation import Segmentation
from ai21.modules.improvements import Improvements
from ai21.modules.question_answering import Answer
from ai21.modules.sagemaker import SageMaker
from ai21.modules.library import Library
from ai21.modules.gec import GEC
from ai21.modules.embed import Embed
from ai21.modules.destination import BedrockDestination, AI21Destination, SageMakerDestination
from ai21.version import __version__


api_key = None
organization = None
application = None
api_version = DEFAULT_API_VERSION
api_host = STUDIO_HOST
timeout_sec = None
num_retries = None
aws_region = None
log_level = 'warning'



DEFAULT_API_VERSION = 'v1'
STUDIO_HOST = 'https://api.ai21.com'

SAGEMAKER_ENDPOINT_KEY = 'sm_endpoint'
DESTINATION_KEY = 'destination'
BEDROCK_HOST_FORMAT = 'bedrock.{region_name}.amazonaws.com'
BEDROCK_URL_FORMAT = f'https://{BEDROCK_HOST_FORMAT}'

SAGEMAKER_MODEL_PACKAGE_NAMES = [
    'j2-light',
    'j2-mid',
    'j2-ultra',
    'gec',
    'contextual-answers',
    'paraphrase',
    'summarize',
]


class BedrockModelID:
    J2_GRANDE_INSTRUCT = 'ai21.j2-grande-instruct'
    J2_JUMBO_INSTRUCT = 'ai21.j2-jumbo-instruct'
    J2_ULTRA = 'ai21.j2-ultra'
    J2_MID = 'ai21.j2-mid'


MODEL_ID_REMAPPING = {
    BedrockModelID.J2_JUMBO_INSTRUCT: BedrockModelID.J2_ULTRA,
    BedrockModelID.J2_GRANDE_INSTRUCT: BedrockModelID.J2_MID,
}



from ai21.modules.resources.deletable_resource import DeletableResource
from ai21.modules.resources.execution_utils import execute_studio_request
from ai21.utils import validate_mandatory_field

from ai21.modules.resources.nlp_task import NLPTask

from ai21.modules.resources.listable_resource import ListableResource
from ai21.modules.resources.retrievable_resource import RetrievableResource
from ai21.modules.resources.updatable_resource import UpdatableResource
from ai21.modules.resources.uploadable_resource import UploadableResource


class Library:
    class Files(UploadableResource, RetrievableResource, ListableResource, UpdatableResource, DeletableResource):
        MODULE_NAME = 'library/files'

        @classmethod
        def upload(cls, file_path: str, **params):
            return super().upload(file_path=file_path, file_param_name='file', **params)

    class Answer(NLPTask):
        MODULE_NAME = 'library/answer'

        @classmethod
        def _get_call_name(cls) -> str:
            return cls.MODULE_NAME

        @classmethod
        def _validate_params(cls, params):
            validate_mandatory_field(
                key='question', call_name=cls.MODULE_NAME, params=params, validate_type=True, expected_type=str)

        @classmethod
        def _execute_studio_api(cls, params):
            url = cls.get_base_url(**params)
            url = f'{url}/{cls.MODULE_NAME}'
            return execute_studio_request(task_url=url, params=params)

    class Search(NLPTask):
        MODULE_NAME = 'library/search'

        @classmethod
        def _get_call_name(cls) -> str:
            return cls.MODULE_NAME

        @classmethod
        def _validate_params(cls, params):
            validate_mandatory_field(
                key='query', call_name=cls.MODULE_NAME, params=params, validate_type=True, expected_type=str)

        @classmethod
        def _execute_studio_api(cls, params):
            url = cls.get_base_url(**params)
            url = f'{url}/{cls.MODULE_NAME}'
            return execute_studio_request(task_url=url, params=params)



from ai21.modules.resources.creatable_resource import CreatableResource
from ai21.modules.resources.listable_resource import ListableResource
from ai21.modules.resources.retrievable_resource import RetrievableResource
from ai21.utils import validate_mandatory_field


class CustomModel(CreatableResource, RetrievableResource, ListableResource):
    MODULE_NAME = 'custom-model'

    @classmethod
    def create(cls, **params):
        validate_mandatory_field(key='dataset_id', call_name=cls.MODULE_NAME, params=params, validate_type=True, expected_type=str)
        validate_mandatory_field(key='model_name', call_name=cls.MODULE_NAME, params=params, validate_type=True, expected_type=str)
        validate_mandatory_field(key='model_type', call_name=cls.MODULE_NAME, params=params, validate_type=True, expected_type=str)
        super().create(**params)



from abc import ABCMeta, abstractmethod

from ai21.constants import SAGEMAKER_ENDPOINT_KEY, DESTINATION_KEY
from ai21.errors import WrongInputTypeException
from ai21.modules.destination import Destination, SageMakerDestination, AI21Destination, BedrockDestination
from ai21.modules.resources.ai21_module import AI21Module
from ai21.utils import extract_destination, raise_unsupported_destination


class NLPTask(AI21Module, metaclass=ABCMeta):
    @classmethod
    def execute(cls, **params):
        cls._validate_params(params)

        destination = cls._resolve_destination(params)

        if isinstance(destination, BedrockDestination):
            return cls._execute_bedrock_api(destination=destination, params=params)
        if isinstance(destination, SageMakerDestination):
            return cls._execute_sm(destination=destination, params=params)
        if isinstance(destination, AI21Destination):
            return cls._execute_studio_api(params)

        raise WrongInputTypeException(key=DESTINATION_KEY, expected_type=Destination, given_type=type(destination))

    @classmethod
    @abstractmethod
    def _get_call_name(cls) -> str:
        pass

    @classmethod
    @abstractmethod
    def _validate_params(cls, params):
        pass

    @classmethod
    def _resolve_destination(cls, params) -> Destination:
        sm_endpoint = params.pop(SAGEMAKER_ENDPOINT_KEY, None)

        destination = extract_destination(params) or AI21Destination()
        if sm_endpoint:
            destination = SageMakerDestination(endpoint_name=sm_endpoint)

        return destination

    @classmethod
    def _execute_sm(cls, destination: SageMakerDestination, params):
        raise_unsupported_destination(destination=destination, call_name=cls._get_call_name())

    @classmethod
    def _execute_bedrock_api(cls, destination: BedrockDestination, params):
        raise_unsupported_destination(destination=destination, call_name=cls._get_call_name())

    @classmethod
    @abstractmethod
    def _execute_studio_api(cls, params):
        pass



import json
from typing import Optional, Dict

import requests
from requests.adapters import HTTPAdapter, Retry, RetryError

from ai21.errors import BadRequest, Unauthorized, UnprocessableEntity, TooManyRequests, ServerError, ServiceUnavailable, \
    APIError
from ai21.utils import log_info, log_error

DEFAULT_TIMEOUT_SEC = 300
DEFAULT_NUM_RETRIES = 0
RETRY_BACK_OFF_FACTOR = 0.5
TIME_BETWEEN_RETRIES = 1000
RETRY_ERROR_CODES = (429, 500, 503)
RETRY_METHOD_WHITELIST = ['GET', 'POST', 'PUT']


def handle_non_success_response(status_code: int, response_text: str):
    if status_code == 400:
        raise BadRequest(details=response_text)
    if status_code == 401:
        raise Unauthorized(details=response_text)
    if status_code == 422:
        raise UnprocessableEntity(details=response_text)
    if status_code == 429:
        raise TooManyRequests(details=response_text)
    if status_code == 500:
        raise ServerError(details=response_text)
    if status_code == 503:
        raise ServiceUnavailable(details=response_text)
    raise APIError(status_code, details=response_text)


def requests_retry_session(session, retries=0):
    retry = Retry(total=retries, read=retries, connect=retries, backoff_factor=RETRY_BACK_OFF_FACTOR,
                  status_forcelist=RETRY_ERROR_CODES, method_whitelist=frozenset(RETRY_METHOD_WHITELIST))
    adapter = HTTPAdapter(max_retries=retry)
    session.mount('https://', adapter)
    session.mount('http://', adapter)
    return session


class HttpClient:
    def __init__(self, timeout_sec: int = None, num_retries: int = None, headers: Dict = None):
        self.timeout_sec = timeout_sec if timeout_sec is not None else DEFAULT_TIMEOUT_SEC
        self.num_retries = num_retries if num_retries is not None else DEFAULT_NUM_RETRIES
        self.headers = headers if headers is not None else {}
        self.apply_retry_policy = self.num_retries > 0

    def execute_http_request(self, method: str, url: str, params: Optional[Dict] = None, files=None, auth=None):
        session = requests_retry_session(requests.Session(),
                                         retries=self.num_retries) if self.apply_retry_policy else requests.Session()
        timeout = self.timeout_sec
        headers = self.headers
        data = json.dumps(params).encode()
        log_info(f'Calling {method} {url} {headers} {data}')
        try:
            if method == 'GET':
                response = session.request(method, url, headers=headers, timeout=timeout)
            elif files is not None:
                if method != 'POST':
                    raise ValueError(
                        f'execute_http_request supports only POST for files upload, but {method} was supplied instead')
                if 'Content-Type' in headers:
                    headers.pop(
                        'Content-Type')  # multipart/form-data 'Content-Type' is being added when passing rb files and payload
                response = session.request(method, url, headers=headers, data=params, files=files, timeout=timeout,
                                           auth=auth)
            else:
                response = session.request(method, url, headers=headers, data=data, timeout=timeout, auth=auth)
        except ConnectionError as connection_error:
            log_error(f'Calling {method} {url} failed with ConnectionError: {connection_error}')
            raise connection_error
        except RetryError as retry_error:
            log_error(f'Calling {method} {url} failed with RetryError after {self.num_retries} attempts: {retry_error}')
            raise retry_error
        except Exception as exception:
            log_error(f'Calling {method} {url} failed with Exception: {exception}')
            raise exception

        if response.status_code != 200:
            log_error(f'Calling {method} {url} failed with a non-200 response code: {response.status_code}')
            handle_non_success_response(response.status_code, response.text)

        return response.json()



from ai21.modules.resources.execution_utils import execute_studio_request
from ai21.modules.resources.nlp_task import NLPTask
from ai21.utils import validate_mandatory_field


class Embed(NLPTask):
    MODULE_NAME = 'embed'

    @classmethod
    def _get_call_name(cls) -> str:
        return cls.MODULE_NAME

    @classmethod
    def _validate_params(cls, params):
        validate_mandatory_field(key='texts', call_name=cls.MODULE_NAME, params=params, validate_type=True,
                                 expected_type=list)

    @classmethod
    def _execute_studio_api(cls, params):
        url = cls.get_base_url(**params)
        url = f'{url}/{cls.MODULE_NAME}'
        return execute_studio_request(task_url=url, params=params)


from ai21.modules.destination import SageMakerDestination
from ai21.modules.resources.ai21_module import AI21Module
from ai21.modules.resources.execution_utils import execute_studio_request, execute_sm_request
from ai21.modules.resources.nlp_task import NLPTask
from ai21.utils import validate_mandatory_field


class Experimental(AI21Module):

    @classmethod
    def embed(cls, **params):
        validate_mandatory_field(key='texts', call_name="embed_experimental", params=params, validate_type=True,
                                 expected_type=list)
        url = cls.get_base_url(**params)
        url = f'{url}/experimental/embed'
        return execute_studio_request(task_url=url, params=params)

    @classmethod
    def answer(cls, **params):
        return _AnswerExperimental.execute(**params)


class _AnswerExperimental(NLPTask):

    MODULE_NAME = 'answer_experimental'

    @classmethod
    def _get_call_name(cls) -> str:
        return cls.MODULE_NAME

    @classmethod
    def _validate_params(cls, params):
        validate_mandatory_field(key='context', call_name="answer_experimental", params=params, validate_type=True,
                                 expected_type=str)
        validate_mandatory_field(key='question', call_name="answer_experimental", params=params, validate_type=True,
                                 expected_type=str)

    @classmethod
    def _execute_sm(cls, destination: SageMakerDestination, params):
        return execute_sm_request(destination=destination, params=params)

    @classmethod
    def _execute_studio_api(cls, params):
        url = cls.get_base_url(**params)
        url = f'{url}/experimental/answer'
        return execute_studio_request(task_url=url, params=params)
